use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007004
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
Compiler
DROLSKY
DROLSKY's
Dave
Exceptions
Gregory
Konojacki
Oschwald
Params
PayPal
Rolsky
Rolsky's
Tomasz
ValidationCompiler
autarch
drolsky
goschwald
lib
me
oschwald
params
slurpy
uncompromised
validator
validators
